# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/02/08
#
"""
实现 transformer 的 encode 部分
"""
from torch import nn

from pure_attention.base_transformer.layers import HalfFeedForward, AddNorm, AttentionAddNorm


class EncoderLayer(nn.Module):
    """
    注意这个结构和 bert 的结构完全一致, 在 bert 中被称作 BertLayer。
    """

    def __init__(self, config):
        super(EncoderLayer, self).__init__()
        # Multi-Head Attention
        self.attention = AttentionAddNorm(config)

        # Feed Forward + Add & Norm
        self.intermediate = HalfFeedForward(config)
        self.output = AddNorm(config.intermediate_size, config.hidden_size,
                              config.hidden_dropout_prob, config.layer_norm_eps)

    def forward(self, query_tensor, key_tensor, value_tensor, attention_mask=None):
        # Multi-Head Attention
        attention_output = self.attention(query_tensor, key_tensor, value_tensor, attention_mask)

        # Feed Forward + Add & Norm
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)

        return layer_output


class Encoder(nn.Module):
    def __init__(self, config):
        super(Encoder, self).__init__()
        # PyTorch 中的 ModuleList https://zhuanlan.zhihu.com/p/64990232
        self.layer = nn.ModuleList([EncoderLayer(config) for _ in range(config.num_hidden_layers)])

    def forward(self, hidden_states, attention_mask=None):
        for layer_module in self.layer:
            # encode 部分是 self-attention ，qkv 来源一致
            hidden_states = layer_module(hidden_states, hidden_states, hidden_states, attention_mask)

        return hidden_states



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/01/22
#
""""""
import os

import torch
from torch import nn

from pure_attention.backbone_bert.package import BertConfig, BertOutput
from pure_attention.base_transformer.encoder import Encoder
from pure_attention.base_transformer.layers import InputEmbeddings
from pure_attention.utils.logger import init_logger


class BertPooler(nn.Module):
    def __init__(self, config):
        super(BertPooler, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        # 只取出第一个 token 也就是 cls 位置上的 embedding 进行 dense 变形
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class BertModel(nn.Module):
    def __init__(self, model_path):
        super(BertModel, self).__init__()

        self.config = BertConfig(os.path.join(model_path, "config.json"))  # 配置文件一定要有
        self.logger = init_logger(self.__class__.__name__)

        self.embeddings = InputEmbeddings(self.config)
        self.encoder = Encoder(self.config)
        self.pooler = BertPooler(self.config)

        self.init_weights()
        self.from_pretrained(os.path.join(os.path.join(model_path, "pytorch_model.bin")))
        self.eval()

    def init_weights(self):
        self.apply(self._init_weights)

    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def from_pretrained(self, pretrained_model_path):
        if not os.path.exists(pretrained_model_path):
            print(f"missing pretrained_model_path: {pretrained_model_path}")
            return

        state_dict = torch.load(pretrained_model_path, map_location='cpu')

        # 替换部分变量名
        for old_key in state_dict.copy().keys():
            new_key = old_key
            if 'gamma' in old_key:
                new_key = new_key.replace('gamma', 'weight')
            if 'beta' in old_key:
                new_key = new_key.replace('beta', 'bias')
            if 'bert.' in old_key:
                new_key = new_key.replace('bert.', '')
            # 兼容部分不优雅的变量命名
            if 'LayerNorm' in old_key:
                new_key = new_key.replace('LayerNorm', 'layer_norm')

            if new_key != old_key:
                state_dict[new_key] = state_dict.pop(old_key)

        missing_keys, unexpected_keys = self.load_state_dict(state_dict, strict=False)
        # 可能存在少许不影响结果的参数
        if len(missing_keys):
            self.logger.warning("\n\t".join(["missing_keys:"] + missing_keys))
        if len(unexpected_keys):
            self.logger.warning("\n\t".join(["unexpected_keys:"] + unexpected_keys))

    def forward(self, input_ids, attention_mask=None, token_type_ids=None, position_ids=None):
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        # We create a 4D attention mask from a 2D tensor mask.
        # Sizes are [batch_size, 1, 1, to_seq_length]
        # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]

        # 使用 dataloader 的时候纬度可能出问题 todo @mmmwhy
        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype)  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        embedding_output = self.embeddings(input_ids, position_ids=position_ids, token_type_ids=token_type_ids)
        sequence_output = self.encoder(embedding_output, extended_attention_mask)
        pooled_output = self.pooler(sequence_output)

        outputs = BertOutput(last_hidden_state=sequence_output, pooler_output=pooled_output)

        return outputs



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/02/08
#
"""
完整的 transformer 结构
"""
from torch import nn

from pure_attention.base_transformer.decoder import Decoder
from pure_attention.base_transformer.encoder import Encoder
from pure_attention.base_transformer.layers import InputEmbeddings


class Transformer(nn.Module):
    def __init__(self, config):
        super(Transformer, self).__init__()
        self.encoder = Encoder(config)
        self.decoder = Decoder(config)

        self.src_embed = InputEmbeddings(config, config.src_vocab_size)
        self.tgt_embed = InputEmbeddings(config, config.tgt_vocab_size)

        self.proj = nn.Linear(config.hidden_size, config.tgt_vocab_size)

    def encode(self, src, src_mask):
        return self.encoder(self.src_embed(src), src_mask)

    def decode(self, enc_output, src_mask, tgt, tgt_mask):
        return self.decoder(self.tgt_embed(tgt), enc_output, src_mask, tgt_mask)

    def forward(self, src, tgt, src_mask, tgt_mask):
        # encoder 的结果作为 decoder 的 enc_output 参数传入，进行 decode
        dec_output = self.decode(self.encode(src, src_mask), src_mask, tgt, tgt_mask)

        # decode 后的结果，先进入一个全连接层变为词典大小的向量，然后进行 log_softmax 操作 (在 softmax 结果上再做多一次 log 运算)
        # log_softmax 与 softmax 的区别在哪里？ https://www.zhihu.com/question/358069078
        return nn.LogSoftmax(dim=-1)(self.proj(dec_output))



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/02/12
#
""""""
import math

import torch
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR


def get_cosine_schedule_with_warmup(
    optimizer: Optimizer, num_warmup_steps: int, num_training_steps: int, num_cycles: float = 0.5, last_epoch: int = -1
):
    """
    搬运自：https://github.com/huggingface/transformers/blob/2e9af294940083915ccb2740a7c8d5b154194f15/src/transformers/optimization.py#L103-L134
    优化曲线可以参考：https://cdn.lifeiyang.cn/img/202202121615356.png

    Create a schedule with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, after a warmup period during which it increases linearly between 0 and the
    initial lr set in the optimizer.
    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        num_cycles (:obj:`float`, `optional`, defaults to 0.5):
            The number of waves in the cosine schedule (the defaults is to just decrease from the max value to 0
            following a half-cosine).
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.
    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return LambdaLR(optimizer, lr_lambda, last_epoch)



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/01/25
#
""""""
import json

import torch


class BertConfig:
    def __init__(self, config_json_file):
        """
        定制化的 config，__getattr__ 处进行判断，可以方便的进行 config.xxx 操作，并兼容不存在的情况。
        """
        with open(config_json_file, "r", encoding='utf-8') as reader:
            json_config = json.loads(reader.read())
        for key, value in json_config.items():
            self.__dict__[key] = value

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        return None


class BertOutput:
    def __init__(self, last_hidden_state: torch.FloatTensor, pooler_output: torch.FloatTensor):
        self.last_hidden_state = last_hidden_state
        self.pooler_output = pooler_output


class TokenizerOutput:
    def __init__(self, token_ids, segment_ids=None, attention_mask=None):
        # 以下写法是等价的
        self.input_ids = token_ids
        self.token_type_ids = segment_ids
        self.attention_mask = attention_mask



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/02/07
#
"""
实现 transformer 的 decode 部分
"""
from torch import nn

from pure_attention.base_transformer.layers import HalfFeedForward, AddNorm, AttentionAddNorm


class DecoderLayer(nn.Module):
    def __init__(self, config):
        super(DecoderLayer, self).__init__()
        # Masked Multi-head Attention
        self.self_attn = AttentionAddNorm(config)

        # Multi-Head Attention
        self.enc_attn = AttentionAddNorm(config)

        # Feed Forward + Add & Norm
        self.intermediate = HalfFeedForward(config)
        self.output = AddNorm(config.intermediate_size, config.hidden_size,
                              config.hidden_dropout_prob, config.layer_norm_eps)

    def forward(self, dec_input, enc_output, slf_attn_mask, dec_enc_attn_mask):
        # Masked Multi-head Attention
        dec_self_attn = self.self_attn(dec_input, dec_input, dec_input, dec_enc_attn_mask)

        # Multi-Head Attention
        dec_enc_attn = self.enc_attn(dec_self_attn, enc_output, enc_output, slf_attn_mask)

        # Feed Forward + Add & Norm
        intermediate_output = self.intermediate(dec_enc_attn)
        layer_output = self.output(intermediate_output, dec_enc_attn)

        return layer_output


class Decoder(nn.Module):
    def __init__(self, config):
        super(Decoder, self).__init__()
        # PyTorch 中的 ModuleList https://zhuanlan.zhihu.com/p/64990232
        self.layer = nn.ModuleList([DecoderLayer(config) for _ in range(config.num_hidden_layers)])

    def forward(self, dec_input, enc_output, slf_attn_mask, dec_enc_attn_mask):
        for layer_module in self.layer:
            dec_input = layer_module(dec_input, enc_output, slf_attn_mask, dec_enc_attn_mask)
        return dec_input



# !/usr/bin/python
# -*- coding: utf-8 -*-
#
# @author: mmmwhy <mmmwhy@mail.ustc.edu.cn>
# @date: 2022/01/22
#

import json

from setuptools import find_packages, setup


# 读取文件内容
def read_file(filename):
    with open(filename, encoding='utf-8') as f:
        long_description = f.read()
    return long_description


# 获取依赖
def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]


setup(
    name='pure_attention',
    version=json.loads(read_file("package.json"))["version"],
    description='use pure attention implement cv/nlp backbone',
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    license='Apache License 2.0',
    url="https://github.com/mmmwhy/pure_attention",
    author='mmmwhy',
    author_email="mmmwhy@mail.ustc.edu.cn",
    install_requires=read_requirements("requirements.txt"),
    packages=find_packages(include=['pure_attention.*']),
    include_package_data=True
)
# setup(
#     name='pure_attention',
#     version="0.0.17",
#     description='use pure attention implement cv/nlp backbone',
#     long_description="",
#     long_description_content_type="text/markdown",
#     license='Apache License 2.0',
#     url="https://github.com/mmmwhy/pure_attention",
#     author='mmmwhy',
#     author_email="mmmwhy@mail.ustc.edu.cn",
#     install_requires="",
#     packages=["pure_attention"],
#     include_package_data=True
# )
